
c
c Neighbourhood-search calling routine for the Raysum ray-theoretical
c teleseismic code.
c
      Program raysum_na

c					MEMORY AND ARRAY SIZES
c
c					The NA routines use the
c					following include file to define
c					all parameters controlling
c					memory required by the arrays.
c
c					A description of each parameter,
c					indicating which ones should be 
c					changed to fit your application
c					can be found in 'na_param.inc' 
c
c	include 	'na_param.inc'
c
c					INPUT AND OUTPUT
c       Input files:
c                   na.in       	Contains options for Neighbourhood 
c					algorithm (see HTML manual)
c
c       Output files:
c                   na.sum      	summary of results
c
c                   na.nad      	NAD (NA direct access file of models)
c
c                   sobol.coeff 	initializing data used for
c                               	quasi-random sequences
c					(This is output for reference only)
c
c       Comments:
c                The NAD file is a direct access compact format file
c                containing all models generated by the neighbourhood
c                algorithm. (NAD files can be read in by multi-dimensional
c                integration program NA-Bayes and plot program S-plot.)
c
c		 Logical units 30-40 are reserved for use by NA subroutines
c		 The user specific routines should not use these values
c		 for logical units.
c		 The NA routines also write to LU 6 as standard out.
c
c ----------------------------------------------------------------------------
c
c					Call NA routine to do the work		
        call na

      end
c
c ----------------------------------------------------------------------------
c
c Raysum na-interfacing routines: 
c
c	user_init - defines parametrization and can be
c		    used to perform any user initialization 
c		    (e.g. reading in data etc.) 
c
c	forward   - performs forward modelling on an input model
c		    and returns the misfit value of that model.
c
c	writemodels - a routine called after the NA has completed.
c		      All models and misfits are supplied 
c		      to the routine and here the user may write  
c		      then out in their own format.
c		      (Note that the NA routines only write the
c		       models out in a direct access NAD file.
c		       see HTML manual for details.)
c
c
c
c-------------------------------------------------------------------------
c
c	user_init - performs all user specific initialization tasks. 
c
c	Input: - none
c
c	Output:
c	      nd		:Number of dimensions in parameter space
c	      ranges(2,nd)	:bounds on parameter space
c	      scales(nd+1)	:scale factors in parameter space
c
c	Comments:
c		 The scale factors are used to transform the model
c		 parameter space into a non-dimensional space. They 
c		 must be provided by the user to put each parameter
c		 on an equal footing.  In the transformation
c		 the i-th parameter becomes,
c		
c		 xi_after = (xi_before-ranges(1,i))/scales(i+1).
c
c		 Note that scales(i+1) contains the scale factor for 
c		 the i-th parameter. The value of scales(1) can be used 
c		 as a shortcut to indicate common choices for scale 
c		 factors. If scales(1) = 0 then all scale factors are
c		 reset to 1; if scales(1) = -1 then the scale factors 
c		 are reset to the parameter range, i.e.,
c
c		 scales(i+1) = ranges(2,i)-ranges(1,i)
c
c		 and for any other value the scale factors are left unchanged.
c		 Note that if scales(1) is set to 0 or -1 then there is
c		 no need to fill up the scales array.
c
c		 This routine is only set up for a 6 layer 24 dimensional 
c		 receiver function problem. 
c
c				M. Sambridge, RSES (ANU), April 1998. 
c
c-------------------------------------------------------------------------
c
      subroutine user_init(nd,ranges,scales)

c           Initialize raysum forward modelling

        implicit none
        include PARAMFILE
        include 'na_param.inc'
        
        integer nd
        real*4 ranges(2,*),scales(*)


c Model parameters -- we're going to use a single matrix for all real
c params -- mod_min and mod_max are dimensioned (layer,param) where
c parameters are: 1) thickness, 2) rho, 3) alpha, 4) beta, 5) pct_a,
c 6) pct_b, 7) trend, 8) plunge, 9) strike, 10) dip
c The corresponding mod_flags indicates which model parameters are
c allowed to vary -- and are to be 'packed' into the NA model vector.
c If mod_flag for a particular parameter is false, mod_min is used.
        integer nlay
        real mod_min(maxlay,10),mod_max(maxlay,10)
        logical mod_flag(maxlay,10),isoflag(maxlay)
c Geometry parameters
        real baz(maxtr),slow(maxtr),sta_dx(maxtr),sta_dy(maxtr)
        integer ntr
c Phase parameters
        integer phaselist(maxseg,2,maxph),nseg(maxph),numph,iphase
c Data parameters
        real data_tr(3,maxsamp,maxtr),dt,shift,weight(maxtr)
        integer nsamp,align
c Parameters from parameter file
        real width,cweight(3)
        integer mults,out_rot,misfit_type

c            Common block for raysum calling routines
        common /raysum_com/nlay,mod_min,isoflag,mod_flag,baz,slow,
     &                     sta_dx,sta_dy,ntr,phaselist,nseg,numph,
     &                     iphase,data_tr,dt,shift,nsamp,align,weight,
     &                     mults,width,out_rot,cweight,misfit_type
     
     
c Scratch variables
        character modname*(namelen),phname*(namelen)
        integer nlay2,ilay,iparam,j
        
        modname='model_max'
        call readmodel(modname,mod_max(1,1),mod_max(1,2),mod_max(1,3),
     &                 mod_max(1,4),isoflag,mod_max(1,5),mod_max(1,6),
     &                 mod_max(1,7),mod_max(1,8),mod_max(1,9),
     &                 mod_max(1,10),nlay2)        
        modname='model_min'
        call readmodel(modname,mod_min(1,1),mod_min(1,2),mod_min(1,3),
     &                 mod_min(1,4),isoflag,mod_min(1,5),mod_min(1,6),
     &                 mod_min(1,7),mod_min(1,8),mod_min(1,9),
     &                 mod_min(1,10),nlay)
     
        if (nlay2 .ne. nlay) then
          write (*,*) 'ERROR -- number of layers doesn''t match for'
          write (*,*) 'minimum and maximum models!'
          stop
        end if
        
c          Shuffle parameters into format NA wants -- use modflag to
c          keep track of which parameters are used.
        nd=0
        do ilay=1,nlay
          do iparam=1,10
            mod_flag(ilay,iparam) = (mod_min(ilay,iparam) .ne.
     &                               mod_max(ilay,iparam))
            if (mod_flag(ilay,iparam)) then
              nd=nd+1
              if (nd .gt. nd_max) then
                write (*,*) 'ERROR -- too many variable parameters!'
                stop
              else
                ranges(1,nd)=mod_min(ilay,iparam)
                ranges(2,nd)=mod_max(ilay,iparam)
              end if              
            end if
          end do
        end do
        
        scales(1)=-1
        
        write (*,*) 'Reading parameters...'
        modname='raysum-na-params'
        call read_inv_params(modname,mults,width,out_rot,cweight,
     &                       misfit_type)
        
        write (*,*) 'Model prep complete'
        write (*,*)
        write (*,*) 'Min/max models:'
        call writemodel(6,mod_min(1,1),mod_min(1,2),mod_min(1,3),
     &                 mod_min(1,4),isoflag,mod_min(1,5),mod_min(1,6),
     &                 mod_min(1,7),mod_min(1,8),mod_min(1,9),
     &                 mod_min(1,10),nlay)
        call writemodel(6,mod_max(1,1),mod_max(1,2),mod_max(1,3),
     &                 mod_max(1,4),isoflag,mod_max(1,5),mod_max(1,6),
     &                 mod_max(1,7),mod_max(1,8),mod_max(1,9),
     &                 mod_max(1,10),nlay2)
        write (*,*) 'Model variability:'
        do ilay=1,nlay
          write (*,'(10(X,L))') mod_flag(ilay,1),mod_flag(ilay,2),
     &         mod_flag(ilay,3),mod_flag(ilay,4),mod_flag(ilay,5),
     &         mod_flag(ilay,6),mod_flag(ilay,7),mod_flag(ilay,8),
     &         mod_flag(ilay,9),mod_flag(ilay,10)
        end do
        write (*,*)
        
        write (*,*) 'Reading geometry'
        modname='geometry'
        call readgeom(modname,baz,slow,sta_dx,sta_dy,ntr)        
        
        if (mults .ne. 3) then
          write (*,*) 'Building phaselist'
          numph=0
          iphase=1
          call ph_direct(phaselist,nseg,numph,nlay,iphase)
          if (mults .eq. 1) then
            call ph_fsmults(phaselist,nseg,numph,nlay,1,iphase)
          end if
          if (mults .eq. 2) then
            do j=1,nlay-1
              call ph_fsmults(phaselist,nseg,numph,nlay,j,iphase)
            end do
          end if
          call writephases(6,phaselist,nseg,numph)
        else
          write (*,*) 'Reading phaselist from file phase.list'
          phname='phase.list'
          call readphases(phname,phaselist,nseg,numph)
        end if
        
        write (*,*) 'Reading data...'
        modname='data_in'
        call readtraces(modname,data_tr,ntr,nsamp,dt,align,shift)
        write (*,*) 'Got ',ntr,' traces, ',nsamp,' samples.'
        
        write (*,*) 'Reading weights...'
        modname='trace_weight'
        call readweight(modname,weight,ntr)
        
      end
c
c-------------------------------------------------------------------------
c
c	forward - performs forward modelling for user supplied problem.
c		  In this case it calculates predicted receiver functions
c		  for a single model and calculates the misfit measure
c		  between observation and prediction.
c
c	Input: 
c	      nd		:Number of dimensions in parameter space
c	      model(nd)		:input velocity model
c
c	Output:
c	      lppd		:negative log(ppd)
c
c	Comments:
c		 This routine allows the user to perform the forward 
c		 modelling and define an a posterior probability density
c		 function using the resulting mismatch with the observed
c		 data. Note that the output quanity is the negative
c		 natural log of the ppd. This is equivalent to the 
c		 misfit function in most problems. 
c
c				M. Sambridge, RSES (ANU), April 1998. 
c
c-------------------------------------------------------------------------
c
      subroutine forward(nd,model,lppd)
c      Forward-model using raysum; return misfit from real data. 

        implicit none
        include PARAMFILE

        integer nd
	real*4 lppd,model(nd)
        
c Model parameters
        integer nlay
        real mod_min(maxlay,10)
        logical mod_flag(maxlay,10),isoflag(maxlay)
c Geometry parameters
        real baz(maxtr),slow(maxtr),sta_dx(maxtr),sta_dy(maxtr)
        integer ntr
c Phase parameters
        integer phaselist(maxseg,2,maxph),nseg(maxph),numph,iphase
c Data parameters
        real data_tr(3,maxsamp,maxtr),dt,shift,weight(maxtr)
        integer nsamp,align
        
        real width,cweight(3)
        integer mults,out_rot,misfit_type

c            Common block for raysum calling routines
        common /raysum_com/nlay,mod_min,isoflag,mod_flag,baz,slow,
     &                     sta_dx,sta_dy,ntr,phaselist,nseg,numph,
     &                     iphase,data_tr,dt,shift,nsamp,align,weight,
     &                     mults,width,out_rot,cweight,misfit_type
     
c            Scratch variables
        real curmod(maxlay,10),amp_in
        integer ilay,ipar,id
        real aa(3,3,3,3,maxlay),ar_list(3,3,3,3,maxlay,2)
        real rot(3,3,maxlay),tt(maxph,maxtr),amp(3,maxph,maxtr)
        real synth_cart(3,maxsamp,maxtr),synth_ph(3,maxsamp,maxtr)
        
c            Functions
        real getmisfit
     
c            Build curmod from model
        id=0
        do ilay=1,nlay
          do ipar=1,10
            if (mod_flag(ilay,ipar)) then
              id=id+1
              curmod(ilay,ipar)=model(id)
            else
              curmod(ilay,ipar)=mod_min(ilay,ipar)
            end if
          end do
        end do
        
c            Build a_ijkl tensors
        call buildmodel(aa,ar_list,rot,curmod(1,1),curmod(1,2),
     &        curmod(1,3),curmod(1,4),isoflag,curmod(1,5),curmod(1,6),
     &        curmod(1,7),curmod(1,8),curmod(1,9),curmod(1,10),nlay)
     
c            Get arrival times and amplitudes
        amp_in=1.
        call get_arrivals(tt,amp,curmod(1,1),curmod(1,2),isoflag,
     &        curmod(1,9),curmod(1,10),aa,ar_list,rot,baz,slow,sta_dx,
     &        sta_dy,phaselist,ntr,nseg,numph,nlay,amp_in)
     
c            Normalize amplitudes by direct P
        call norm_arrivals(amp,baz,slow,curmod(1,3),curmod(1,4),
     &                     curmod(1,2),ntr,numph,1,1)
        

c            Make traces
        call make_traces(tt,amp,ntr,numph,nsamp,dt,width,align,shift,
     &         synth_cart)
        if (out_rot .eq. 1) then
          call rot_traces(synth_cart,baz,ntr,nsamp,synth_ph)
        end if
        if (out_rot .eq. 2) then
          call fs_traces(synth_cart,baz,slow,curmod(1,3),curmod(1,4),
     &                   curmod(1,2),ntr,nsamp,synth_ph)
        end if
        
c             Get misfit
        if (out_rot .eq. 0) then
          lppd=getmisfit(data_tr,synth_cart,weight,ntr,nsamp,
     &                   cweight(1),cweight(2),cweight(3),misfit_type)
        else
          lppd=getmisfit(data_tr,synth_ph,weight,ntr,nsamp,
     &                   cweight(1),cweight(2),cweight(3),misfit_type)

        end if
        write (*,*) 'Misfit type ',misfit_type,' = ',lppd

                
      end
c
c-------------------------------------------------------------------------
c
c	writemodels - user supplied routine to write out models produced
c		      by Neighbourhood algorithm in user's own format.
c
c	Input: 
c	      nd		  :number of dimensions in parameter space
c	      ntot		  :number of models generated by NA 
c	      models(nd,ntot)     :models generated by NA
c	      misfit 		  :array of model misfits (-lppd's)
c	      ns1  		  :initial sample size used by NA
c	      ns2  		  :normal sample size used by NA
c	      itmax  		  :number of iterations
c	      nh_max  		  :maximum length of nad file header 
c
c	Output:         
c	      nh  		  :length of nad file header 
c	      header(nh)  	  :character string containing nad file header
c
c	Comments:
c	         This routine allows the user to write out the models
c		 produced by the NA in their own format. The three 
c		 parameters ns1, ns2 and itmax define the samples sizes
c		 and number of iterations taken by the NA. Initially ns1
c		 random samples are generated, and then itmax sets of ns2
c		 models are generated and so ntot = ns1 + ns2*itmax. 
c
c		 The user may if they wish write their own information 
c		 into the `user reserved' part of NAD file header.
c		 (See example below for details).
c
c				M. Sambridge, RSES (ANU), Oct 1999. 
c
c-------------------------------------------------------------------------
c
      subroutine writemodels
     &             (nd, ntot, models, misfit, ns1, ns2, itmax,
     &              nh_max, nh, header)

        implicit none
        include PARAMFILE
c						NA variables and arrays
        real*4 models(nd,*),misfit(ntot)
        character*(*) header
        integer nd,ntot,ns1,ns2,itmax,nh_max,nh

c Model parameters
        integer nlay
        real mod_min(maxlay,10)
        logical mod_flag(maxlay,10),isoflag(maxlay)
c Geometry parameters
        real baz(maxtr),slow(maxtr),sta_dx(maxtr),sta_dy(maxtr)
        integer ntr
c Phase parameters
        integer phaselist(maxseg,2,maxph),nseg(maxph),numph,iphase
c Data parameters
        real data_tr(3,maxsamp,maxtr),dt,shift,weight(maxtr)
        integer nsamp,align
        
        real width,cweight(3)
        integer mults,out_rot,misfit_type

c            Common block for raysum calling routines
        common /raysum_com/nlay,mod_min,isoflag,mod_flag,baz,slow,
     &                     sta_dx,sta_dy,ntr,phaselist,nseg,numph,
     &                     iphase,data_tr,dt,shift,nsamp,align,weight,
     &                     mults,width,out_rot,cweight,misfit_type
        
        character filename*(namelen)
        integer index(ntot),i,ilay,ipar,id
        real curmod(maxlay,10)
        
        filename='models.out'
        open(unit=iounit1,file=filename,status='unknown')

        
        write(*,*) 'Sorting models...'
c             Sort models in order of increasing misfit
        do i=1,ntot
          index(i)=i
        end do
        call sort(misfit,index,ntot,ntot)
        write(*,*) 'Models sorted.'
        
c              Write out models
        write(iounit1,*) '# ntot, nd, ns1, ns2, itmax, nlay'
        write(iounit1,*) ntot,nd,ns1,ns2,itmax,nlay
        write(iounit1,*)
        do i=1,ntot
          write(iounit1,*) '# --------------------------------'
          write(iounit1,*) 'Model ',i,' misfit ',misfit(index(i))
          id=0
          do ilay=1,nlay
            do ipar=1,10
              if (mod_flag(ilay,ipar)) then
                id=id+1
                curmod(ilay,ipar)=models(id,index(i))
              else
                curmod(ilay,ipar)=mod_min(ilay,ipar)
              end if
            end do
          end do
          call writemodel(iounit1,curmod(1,1),curmod(1,2),curmod(1,3),
     &                    curmod(1,4),isoflag,curmod(1,5),curmod(1,6),
     &                    curmod(1,7),curmod(1,8),curmod(1,9),
     &                    curmod(1,10),nlay)
        end do
        close(unit=iounit1)
        
c            Write out model table, for plotting
        filename='models.table'
        open(unit=iounit1,file=filename,status='unknown')
        do i=1,ntot
          write (iounit1,'(I6,X,G12.5,$)') index(i),misfit(index(i))
          do id=1,nd
            write (iounit1,'(X,G12.5,$)') models(id,index(i))
          end do
          write (iounit1,*)
        end do
        close(unit=iounit1)                        
      end
      

c -----------------------------------------------

      subroutine read_inv_params(filename,mults,width,out_rot,cweight,
     &                           misfit_type)
      
        implicit none
        include PARAMFILE
        
        character filename*(namelen),buffer*(buffsize)
        integer mults,out_rot,misfit_type,ios,eof
        real width,cweight(3)
        
c          Default values

        mults=1
        width=1.
        out_rot=2
        cweight(1)=0.
        cweight(2)=1.
        cweight(3)=1.
        misfit_type=0
        
        open(unit=iounit1,status='old',file=filename,iostat=ios)
        
        if (ios .eq. 0) then
          write(*,*) 'Reading parameters from ',filename
          call getline(iounit1,buffer,eof)
          read (buffer,*) mults
          call getline(iounit1,buffer,eof)
          read (buffer,*) width
          call getline(iounit1,buffer,eof)
          read (buffer,*) out_rot
          call getline(iounit1,buffer,eof)
          read (buffer,*) cweight(1),cweight(2),cweight(3)
          call getline(iounit1,buffer,eof)
          read (buffer,*) misfit_type
        else
          write (*,*) 'Parameter file ',filename,' does not exist.'
          write (*,*) 'Writing default values.'
          close(unit=iounit1)
          open(unit=iounit1,status='unknown',file=filename)
          write(iounit1,*) '# Multiples: 0 for none, 1 for Moho, 2 ',
     &                      'for all first-order, 3 to read file'
          write(iounit1,*) mults
          write(iounit1,*) '# Gaussian pulse width (seconds)'
          write(iounit1,*) width
          write(iounit1,*) '# Rotation to output: 0 is NS/EW/Z, ',
     &                     '1 is R/T/Z, 2 is P/SV/SH'
          write(iounit1,*) out_rot
          write(iounit1,*) '# Misfit weight of components 1,2,3'
          write(iounit1,*) cweight(1),cweight(2),cweight(3)
          write(iounit1,*) '# Type of misfit to use. 0: correlation,'
          write(iounit1,*) '# 1: scaled L1-norm difference,',
     &                     ' 2: scaled L2-norm difference'
          write(iounit1,*) '# 3: unscaled L1, 4: unscaled L2.'
          write(iounit1,*) misfit_type
        end if
        close(unit=iounit1)
      end
